# -*- coding: utf-8 -*-

from math import ceil

import numpy as np

try:
    from osgeo import gdal, osr
except ImportError:
    import gdal, osr

from .gdalio import _fromDataset, _getDataset
from .gdalspatial import _Projection, _Transformer
from .wrapper import array, full

gdal.UseExceptions()
gdal.PushErrorHandler("CPLQuietErrorHandler")

_RESAMPLING = {
    # A documentation would be nice.
    # There seem to be more functions in GDAL > 2
    "average": gdal.GRA_Average,
    "bilinear": gdal.GRA_Bilinear,
    "cubic": gdal.GRA_Cubic,
    "cubicspline": gdal.GRA_CubicSpline,
    "lanczos": gdal.GRA_Lanczos,
    "mode": gdal.GRA_Mode,
    "nearest": gdal.GRA_NearestNeighbour,
    # only available in GDAL > 2
    "max": getattr(gdal, "GRA_Max", None),
    "min": getattr(gdal, "GRA_Min", None),
}


def _warpTo(source, target, func, max_error=0.125):

    if func is None:
        raise TypeError(
            "Resampling method {:} not available in your GDAL version".format(func)
        )

    target = np.atleast_2d(target)
    if target.ndim < source.ndim:
        target = np.broadcast_to(
            target, source.shape[: -len(target.shape)] + target.shape, subok=True
        )

    target = np.ma.array(
        target,
        mask=target.data == target.fill_value,
        dtype=source.dtype,
        copy=True,
        subok=True,
    )

    target[target.mask] = source.fill_value
    target.fill_value = source.fill_value

    out = _getDataset(target, True)

    gdal.Warp(
        out,
        _getDataset(source),
        resampleAlg=_RESAMPLING[func],
        errorThreshold=max_error,
        # geoloc=True,
        # transformerOptions=["SRC_METHOD=NO_GEOTRANSFORM"]
    )

    # gdal.ReprojectImage(
    #     _getDataset(source), out,
    #     None, None,
    #     _RESAMPLING[func],
    #     0.0, max_error)
    # print _fromDataset(out)

    return _fromDataset(out)


def project(grid, proj, cellsize=None, func="nearest", max_error=0.125):
    # type: (GeoArray, _Projection, Optional[float], str, float) -> GeoArray

    bbox = grid.bbox
    proj = _Projection(proj)
    trans = _Transformer(grid.proj, proj)
    uly, ulx = trans(bbox["ymax"], bbox["xmin"])
    lry, lrx = trans(bbox["ymin"], bbox["xmax"])
    ury, urx = trans(bbox["ymax"], bbox["xmax"])
    lly, llx = trans(bbox["ymin"], bbox["xmin"])

    # Calculate cellsize, i.e. same number of cells along the diagonal.
    if cellsize is None:
        src_diag = np.sqrt(grid.nrows**2 + grid.ncols**2)
        trg_diag = np.sqrt((lly - ury) ** 2 + (llx - urx) ** 2)
        cellsize = trg_diag / src_diag

    # number of cells
    ncols = int(
        abs(ceil((max(urx, lrx, ulx, llx) - min(urx, lrx, ulx, llx)) / cellsize))
    )
    nrows = int(
        abs(ceil((max(ury, lry, uly, lly) - min(ury, lry, uly, lly)) / cellsize))
    )

    target = array(
        data=np.full((grid.nbands, nrows, ncols), grid.fill_value, grid.dtype),
        fill_value=grid.fill_value,
        dtype=grid.dtype,
        yorigin=max(uly, ury, lly, lry),
        xorigin=min(ulx, urx, llx, lrx),
        origin="ul",
        ycellsize=abs(cellsize) * -1,
        xcellsize=cellsize,
        # cellsize   = cellsize,
        proj=proj,
        mode=grid.mode,
    )

    out = resample(source=grid, target=target, func=func, max_error=max_error)
    return out.trim()


def resample(source, target, func="nearest", max_error=0.125):
    return _warpTo(source=source, target=target, func=func, max_error=max_error)


def rescale(source, scaling_factor, func="nearest"):
    shape = list(source.shape[:-2]) + [
        int(s / scaling_factor) for s in source.shape[-2:]
    ]
    cellsize = tuple(c * scaling_factor for c in source.cellsize)
    scaled_grid = full(
        shape,
        source.fill_value,
        origin=source.origin,
        xorigin=source.xorigin,
        yorigin=source.yorigin,
        cellsize=cellsize,
        dtype=source.dtype,
    )
    return resample(source, scaled_grid, func=func)



# -*- coding: utf-8 -*-

import warnings

import numpy as np

try:
    from osgeo import gdal, osr
except ImportError:
    import gdal
    import osr

gdal.UseExceptions()
gdal.PushErrorHandler("CPLQuietErrorHandler")


class _Projection(object):
    def __init__(self, arg):
        """
        Arguments:
        arg can be:
        1. int  : EPSG code
        2. dict : pyproj compatable dictionary
        3. str  : WKT string
        4. _Projection
        """
        self._srs = osr.SpatialReference()
        self._import(arg)

    def _import(self, value):
        if isinstance(value, _Projection):
            self._srs = value._srs
        elif value:
            if isinstance(value, int):
                method = self._srs.ImportFromEPSG
            elif isinstance(value, str):
                method = self._srs.ImportFromWkt
            elif isinstance(value, dict):
                try:
                    method = self._srs.ImportFromDict
                except AttributeError:
                    method = self._srs.ImportFromProj4
                    value = "+{:}".format(
                        " +".join(["=".join(map(str, pp)) for pp in value.items()])
                    )
            else:
                raise RuntimeError("Projection not understood")

            if method(value):
                raise RuntimeError("Failed to set projection")

    def __eq__(self, other):
        return bool(self._srs.IsSame(_Projection(other)._srs))

    def toWkt(self):
        return self._srs.ExportToPrettyWkt()

    def toProj4(self):
        return self._srs.ExportToProj4()

    def toDict(self):
        proj = self._srs.ExportToProj4()
        out = dict(
            filter(
                lambda x: len(x) == 2,
                [p.replace("+", "").split("=") for p in proj.split(" +")],
            )
        )
        return out

    def __nonzero__(self):
        # is a an projection set?
        return bool(self.toWkt())

    def __get__(self, *args, **kwargs):
        return self

    def __set__(self, obj, val):
        self._import(val)

    def __str__(self):
        return self.toWkt()

    def __repr__(self):
        return self.toWkt()


class _Transformer(object):
    def __init__(self, sproj, tproj):
        """
        Arguments
        ---------
        sproj, tproj : Projection

        Purpose
        -------
        Encapsulates the osr Cordinate Transformation functionality
        """
        self._tx = osr.CoordinateTransformation(sproj._srs, tproj._srs)

    def __call__(self, y, x):
        try:
            xt, yt, _ = self._tx.TransformPoint(x, y)
        except NotImplementedError:
            raise AttributeError("Projections not correct or given!")
        return yt, xt



# -*- coding: utf-8 -*-

"""
Author
------
David Schaefer

Purpose
-------
This module provides initializer function for core.GeoArray
"""

import numpy as np

from .core import GeoArray
from .gdalio import _fromDataset, _fromFile
from .gdalspatial import _Projection
from .geotrans import _Geolocation, _Geotrans
from .utils import _tupelize

# from typing import Optional, Union, Tuple, Any, Mapping, AnyStr


def array(
    data,  # type: Union[np.ndarray, GeoArray]
    dtype=None,  # type: Optional[Union[AnyStr, np.dtype]]
    yorigin=0,  # type: Optional[float]
    xorigin=0,  # type: Optional[float]
    origin="ul",  # type: Optional[floatAnyStr]
    fill_value=None,  # type: Optional[float]
    cellsize=1,  # type: Optional[float]]
    ycellsize=None,  # type: Optional[float]
    xcellsize=None,  # type: Optional[float]
    yparam=0,  # type: Optional[float]
    xparam=0,  # type: Optional[float]
    # geotrans   = None,  # type: Optional[_Geotrans]
    yvalues=None,  # type: Optional[np.ndarray]
    xvalues=None,  # type: Optional[np.ndarray]
    proj=None,  # type: Mapping[AnyStr, Union[AnyStr, float]]
    mode="r",  # type: AnyStr
    color_mode="L",  # type: AnyStr
    copy=False,  # type: bool
    fobj=None,  # type: Optional[osgeo.gdal.Dataset]
):  # type: (...) -> GeoArray
    """
    Arguments
    ---------
    data         : numpy.ndarray  # data to wrap

    Optional Arguments
    ------------------
    dtype        : str/np.dtype                  # type of the returned grid
    yorigin      : int/float, default: 0         # y-value of the grid's origin
    xorigin      : int/float, default: 0         # x-value of the grid's origin
    origin       : {"ul","ur","ll","lr"},        # position of the origin. One of:
                   default: "ul"                 #     "ul" : upper left corner
                                                 #     "ur" : upper right corner
                                                 #     "ll" : lower left corner
                                                 #     "lr" : lower right corner
    fill_value   : inf/float                     # fill or fill value
    cellsize     : int/float or 2-tuple of those # cellsize, cellsizes in y and x direction
    proj         : dict/None                     # proj4 projection parameters
    copy         : bool                          # create a copy of the given data

    Returns
    -------
    GeoArray

    Purpose
    -------
    Create a GeoArray from data.
    """

    def _checkGeolocArray(array, diffaxis):
        array = np.asarray(array)
        diff = np.diff(array, axis=diffaxis)

        assert array.ndim == 2
        assert array.shape == data.shape[-2:]
        assert len(np.unique(np.sign(diff))) == 1

        return array

    if yvalues is not None and xvalues is not None:
        yvalues = _checkGeolocArray(yvalues, 0)
        xvalues = _checkGeolocArray(xvalues, 1)
        geotrans = _Geolocation(yvalues, xvalues, shape=data.shape, origin=origin)
    else:

        cellsize = _tupelize(cellsize)

        if ycellsize is None:
            ycellsize = cellsize[0]
            if (origin[0] == "u" and ycellsize > 0) or (
                origin[0] == "l" and ycellsize < 0
            ):
                ycellsize *= -1

        if xcellsize is None:
            xcellsize = cellsize[-1]
            if (origin[1] == "r" and xcellsize > 0) or (
                origin[0] == "l" and xcellsize < 0
            ):
                xcellsize *= -1

        # if geotrans is None:
        # NOTE: not to robust...
        geotrans = _Geotrans(
            yorigin=yorigin,
            xorigin=xorigin,
            ycellsize=ycellsize,
            xcellsize=xcellsize,
            yparam=yparam,
            xparam=xparam,
            origin=origin,
            shape=data.shape,
        )

    proj = _Projection(proj)

    if isinstance(data, GeoArray):
        return GeoArray(
            dtype=dtype or data.dtype,
            geotrans=data.geotrans,
            fill_value=fill_value or data.fill_value,
            proj=proj or data.proj,
            mode=mode or data.mode,
            color_mode=color_mode or data.color_mode,
            fobj=data.fobj,
            data=data.data,
        )

    return GeoArray(
        data=np.array(data, dtype=dtype, copy=copy),
        geotrans=geotrans,
        fill_value=fill_value,
        proj=proj,
        mode=mode,
        color_mode=color_mode,
        fobj=fobj,
    )


def zeros(shape, dtype=np.float64, *args, **kwargs):
    """
    Arguments
    ---------
    shape        : tuple          # shape of the returned grid

    Optional Arguments
    ------------------
    see array

    Returns
    -------
    GeoArray

    Purpose
    -------
    Return a new GeoArray of given shape and type, filled with zeros.
    """

    return array(data=np.zeros(shape, dtype), *args, **kwargs)


def ones(shape, dtype=np.float64, *args, **kwargs):
    """
    Arguments
    ---------
    shape        : tuple          # shape of the returned grid

    Optional Arguments
    ------------------
    see array

    Returns
    -------
    GeoArray

    Purpose
    -------
    Return a new GeoArray of given shape and type, filled with ones.
    """

    return array(data=np.ones(shape, dtype), *args, **kwargs)


def full(shape, value, dtype=np.float64, *args, **kwargs):
    """
    Arguments
    ---------
    shape        : tuple          # shape of the returned grid
    fill_value   : scalar         # fille value

    Optional Arguments
    ------------------
    see array

    Returns
    -------
    GeoArray

    Purpose
    -------
    Return a new GeoArray of given shape and type, filled with fill_value.
    """

    return array(data=np.full(shape, value, dtype), *args, **kwargs)


def empty(shape, dtype=np.float64, *args, **kwargs):
    """
    Arguments
    ----------
    shape        : tuple          # shape of the returned grid

    Optional Arguments
    ------------------
    see array

    Returns
    -------
    GeoArray

    Purpose
    -------
    Return a new empty GeoArray of given shape and type
    """

    return array(data=np.empty(shape, dtype), *args, **kwargs)


def _likeArgs(arr):
    if isinstance(arr, GeoArray):
        return arr.header
    return {}


def zeros_like(arr, dtype=None):
    args = _likeArgs(arr)
    return zeros(shape=arr.shape, dtype=dtype or arr.dtype, **args)


def ones_like(arr, dtype=None):
    args = _likeArgs(arr)
    return ones(shape=arr.shape, dtype=dtype or arr.dtype, **args)


def full_like(arr, value, dtype=None):
    args = _likeArgs(arr)
    return full(shape=arr.shape, value=value, dtype=dtype or arr.dtype, **args)


def fromdataset(ds):
    return array(**_fromDataset(ds))


def fromfile(fname, mode="r"):
    """
    Arguments
    ---------
    fname : str  # file name

    Returns
    -------
    GeoArray

    Purpose
    -------
    Create GeoArray from file

    """
    return _fromFile(fname, mode)



# -*- coding: utf-8 -*-
"""
Purpose
=======

basinex is a basin extractor.
"""

try:
    from ._version import __version__
except ModuleNotFoundError:  # pragma: nocover
    # package is not installed
    __version__ = "0.0.0.dev0"

__all__ = ["__version__"]



# -*- coding: utf-8 -*-

"""
Author
------
David Schaefer

Purpose
-------
This module provides a numpy.ma.MaskedArray as a
wrapper around gdal raster functionality

"""
import copy
import os
import warnings

import numpy as np
from numpy.ma import MaskedArray

from .gdalio import _getDataset, _toFile, _writeData
from .gdalspatial import _Projection
from .geotrans import _Geotrans
from .spatial import SpatialMixin
from .utils import _broadcastedMeshgrid, _broadcastTo

# Possible positions of the grid origin
ORIGINS = (
    "ul",  # "ul" -> upper left
    "ur",  # "ur" -> upper right
    "ll",  # "ll" -> lower left
    "lr",  # "lr" -> lower right
)

_METHODS = (
    # comparison
    "__lt__",
    "__le__",
    "__gt__",
    "__ge__",
    "__eq__",
    "__ne__",
    "__nonzero__",
    # unary
    "__neg__",
    "__pos__",
    "__abs__",
    "__invert__",
    # arithmetic
    "__add__",
    "__sub__",
    "__mul__",
    "__div__",
    "__truediv__",
    "__floordiv__",
    "__mod__",
    "__divmod__",
    "__pow__",
    "__lshift__",
    "__rshift__",
    "__and__",
    "__or__",
    "__xor__",  # "__matmul__",
    # arithmetic, in-place
    "__iadd__",
    "__isub__",
    "__imul__",
    "__idiv__",
    "__itruediv__",
    "__ifloordiv__",
    "__imod__",
    "__ipow__",
    "__ilshift__",
    "__irshift__",
    "__iand__",
    "__ior__",
    "__ixor__",  # "__imatmul__"
)


def _checkMatch(func):
    def inner(*args):
        if len({a.proj for a in args if isinstance(a, GeoArray)}) > 1:
            warnings.warn("Incompatible map projections!", RuntimeWarning)
        if len({a.cellsize for a in args if isinstance(a, GeoArray)}) != 1:
            warnings.warn("Incompatible cellsizes", RuntimeWarning)
        if len({a.getCorner("ul") for a in args if isinstance(a, GeoArray)}) != 1:
            warnings.warn("Incompatible origins", RuntimeWarning)
        return func(*args)

    return inner


class GeoArrayMeta(object):
    def __new__(cls, name, bases, attrs):
        for key in _METHODS:
            attrs[key] = _checkMatch(getattr(MaskedArray, key))
        return type(name, bases, attrs)


class GeoArray(SpatialMixin, MaskedArray):
    """
    Arguments
    ----------
    TODO

    Purpose
    -------
    This numpy.ndarray subclass adds geographic context to data.
    A (hopfully growing) number of operations on the data I/O to/from
    different file formats (see the variable gdalfuncs._DRIVER_DICT)
    is supported.

    Restrictions
    ------------
    Adding the geographic information to the data does (at the moment)
    not imply any additional logic. If the shapes of two grids allow
    the succesful execution of a certain operator/function your program
    will continue. It is within the responsability of the user to check
    whether a given operation makes sense within a geographic context
    (e.g. grids cover the same spatial domain, share a common projection,
    etc.) or not
    """

    __metaclass__ = GeoArrayMeta

    def __new__(
        cls,
        data,
        geotrans=None,
        proj=None,
        fill_value=None,
        fobj=None,
        color_mode=None,  # mask=None,
        yvalues=None,
        xvalues=None,
        mode="r",
        *args,
        **kwargs,
    ):

        # NOTE: The mask will always be calculated, even if its
        #       already present or not needed at all...
        mask = (
            np.zeros_like(data, np.bool) if fill_value is None else data == fill_value
        )

        self = MaskedArray.__new__(
            cls, data=data, fill_value=fill_value, mask=mask, *args, **kwargs
        )
        self.unshare_mask()

        self.__dict__["geotrans"] = geotrans
        self.__dict__["proj"] = _Projection(proj)

        self.__dict__["color_mode"] = color_mode
        self.__dict__["mode"] = mode

        self.__dict__["_fobj"] = fobj
        self.__dict__["_yvalues"] = yvalues
        self.__dict__["_xvalues"] = xvalues

        return self

    def __array_finalize__(self, obj):
        if obj is None:
            return
        super(GeoArray, self).__array_finalize__(obj)
        self._update_from(obj)

    def _update_from(self, obj):

        super(GeoArray, self)._update_from(obj)

        self.__dict__["geotrans"] = getattr(obj, "geotrans", None)
        self.__dict__["proj"] = getattr(obj, "proj", None)
        self.__dict__["color_mode"] = getattr(obj, "color_mode", None)
        self.__dict__["mode"] = getattr(obj, "mode", None)
        self.__dict__["_fobj"] = getattr(obj, "_fobj", None)
        self.__dict__["_yvalues"] = getattr(obj, "_yvalues", None)
        self.__dict__["_xvalues"] = getattr(obj, "_xvalues", None)
        self.__dict__["_geolocation"] = getattr(obj, "_geolocation", None)

    @property
    def header(self):
        out = self._getArgs()
        out.update(out.pop("geotrans")._todict())
        del out["data"]
        return out

    def _getShapeProperty(self, idx):
        try:
            return self.shape[idx] or 1
        except IndexError:
            return 1

    @property
    def nbands(self):
        return self._getShapeProperty(-3)

    @property
    def nrows(self):
        return self._getShapeProperty(-2)

    @property
    def ncols(self):
        return self._getShapeProperty(-1)

    @property
    def fobj(self):
        if self._fobj is None:
            self._fobj = _getDataset(self, mem=True)
        return self._fobj

    def getFillValue(self):
        return super(GeoArray, self).get_fill_value()

    def setFillValue(self, value):
        # change fill_value and update mask
        super(GeoArray, self).set_fill_value(value)
        self.mask = self.data == value
        if value != self.fill_value:
            warnings.warn(
                "Data types not compatible. New fill_value is: {:}".format(
                    self.fill_value
                )
            )

    # decorating the methods did not work out...
    fill_value = property(fget=getFillValue, fset=setFillValue)

    # def __getattribute__(self, key):
    #     "Make descriptors work"
    #     v = object.__getattribute__(self, key)
    #     if hasattr(v, '__get__'):
    #         return v.__get__(None, self)
    #     return v

    # def __setattr__(self, key, value):
    #     "Make descriptors work"
    #     try:
    #         object.__getattribute__(self, key).__set__(self, value)
    #     except AttributeError:
    #         object.__setattr__(self, key, value)

    def __setattr__(self, key, value):
        instance = self.geotrans if hasattr(self.geotrans, key) else self
        object.__setattr__(instance, key, value)

    def __getattr__(self, key):
        try:
            # return object.__getattribute__(self.geotrans, key)
            return getattr(self.geotrans, key)
        except AttributeError:
            raise AttributeError("'GeoArray' object has no attribute '{:}'".format(key))

    @property
    def fobj(self):
        if self._fobj is None:
            self._fobj = _getDataset(self, mem=True)
        return self._fobj

    def _getArgs(
        self,
        data=None,
        fill_value=None,
        geotrans=None,
        mode=None,
        color_mode=None,
        proj=None,
        fobj=None,
    ):

        return {
            "data": data if data is not None else self.data,
            "geotrans": geotrans if geotrans is not None else self.geotrans,
            "proj": proj if proj is not None else self.proj,
            "fill_value": fill_value if fill_value is not None else self.fill_value,
            "mode": mode if mode is not None else self.mode,
            "color_mode": color_mode if color_mode is not None else self.color_mode,
            "fobj": fobj if fobj is not None else self._fobj,
        }

    def fill(self, fill_value):
        """
        works similar to MaskedArray.filled(value) but also changes
        the fill_value and returns an GeoArray instance
        """
        return GeoArray(
            self._getArgs(data=self.filled(fill_value), fill_value=fill_value)
        )

    def __copy__(self):
        return GeoArray(**self._getArgs())

    def __deepcopy__(self, memo):
        return GeoArray(**self._getArgs(data=self.data.copy()))

    def __getitem__(self, slc):

        data = MaskedArray.__getitem__(self, slc)

        # empty array
        if data.size == 0 or np.isscalar(data):
            return data

        geotrans = self.geotrans._getitem(slc)

        return GeoArray(**self._getArgs(data=data.data, geotrans=geotrans))

    def flush(self):
        fobj = self._fobj
        if fobj is not None:
            fobj.FlushCache()
            if self.mode == "a":
                _writeData(self)

    def close(self):
        self.__del__()

    def __del__(self):
        # the virtual memory mapping needs to be released BEFORE the fobj
        self.flush()
        self._fobj = None

    def tofile(self, fname):
        _toFile(self, fname)

    def setMask(self, mask):
        from .wrapper import array

        data = np.array(self, copy=True, subok=False)
        data[..., mask] = self.fill_value
        return array(
            data=data,
            yorigin=self.yorigin,
            xorigin=self.xorigin,
            origin=self.origin,
            cellsize=self.cellsize,
            proj=self.proj,
            fill_value=self.fill_value,
            mode=self.mode,
        )



# -*- coding: utf-8 -*-

import logging
import os
import warnings
from argparse import ArgumentParser
from pathlib import Path

import numpy as np
import yaml

from . import __version__
from . import geoarray as ga
from .extractor import extract
from .gauges import matchFlowacc, readGauges
from .netcdf import NcDimDataset
from .netcdf4 import NcDataset
from .wrapper import GridFile, NcFile


def cli():
    parser = initArgparser()
    args = parser.parse_args()

    logging.basicConfig(
        format="%(message)s", level=logging.DEBUG if args.verbose else logging.INFO
    )

    os.chdir(Path(args.cwd))
    logging.debug(f"in directory: {os.getcwd()}")

    config = readConfig(args.input)

    # command line option -n
    line = args.line
    gauges = readGauges(
        config["gauges"], lat_fix=config.get("latitude-size-correction", False)
    )
    if line:
        if line > len(gauges) - 1:
            raise ValueError("Given line number exceeds table row count")
        gauges = gauges[line : line + 1]

    main(config, gauges)


def readConfig(fname):
    with open(fname, "r") as f:
        try:
            out = yaml.safe_load(f)
        except IOError:
            raise
        except Exception:
            raise RuntimeError("Syntactic error in {:}".format(fname))
    return out


def gaugeBasinMask(flowdir, gauge):
    gauge_idx = flowdir.indexOf(gauge.y, gauge.x)

    mask = np.asarray(
        extract(np.array(flowdir, dtype=np.int32, copy=True), *gauge_idx),
        dtype=np.int32,
    )

    mask[mask == 0] = flowdir.fill_value
    out = ga.array(mask, **flowdir.header)
    return out.trim()


def gridBasinMask(gauge):

    with NcDataset(gauge.path) as ncbase:
        # Infer dimension/fill_value information from the file
        var = ncbase.variables[gauge.varname]
        if var.ndim > 2:
            raise RuntimeError("Expected mask should not have more than 2 dimensions")

        y, x = var.dimensions
        # NOTE: would be a good to have y_shift and x_shift
        #        as optional fields in the gauge lut.
        with NcDimDataset(gauge.path, ydim=y, xdim=x, y_shift=0.5, x_shift=0.5) as nc:
            origin = nc.origin
            out = ga.array(
                nc.variables[gauge.varname][:],
                yorigin=nc.bbox["ymin" if origin[0] == "l" else "ymax"],
                xorigin=nc.bbox["xmin" if origin[1] == "l" else "xmax"],
                origin=origin,
                fill_value=var.fill_value,
                cellsize=nc.cellsize,
            )
    return out.setMask(out <= 0)


def openNcFiles(flist):
    out = {}
    flist = flist or ()
    for fdict in flist:
        fitem = NcFile(**fdict)
        out[fitem] = NcDimDataset(
            fitem.fname,
            "r",
            fitem.ydim,
            fitem.xdim,
            y_shift=fitem.y_shift,
            x_shift=fitem.x_shift,
        )

    return out


def openGridFiles(flist):
    out = {}
    flist = flist or ()
    for fdict in flist:
        out[GridFile(**fdict)] = ga.fromfile(fdict["fname"])
    return out


def writeFiles(bpath, fdict):
    for fitem, fobj in fdict.items():
        path = os.path.join(bpath, fitem.outpath or "")
        if not os.path.isdir(path):
            os.makedirs(path)
        fname = os.path.join(path, os.path.split(fitem.fname)[-1])
        logging.debug("writing file: %s", fname)
        fobj.tofile(fname)


def enlargeFiles(fdict, bbox):
    out = {}
    for fitem, fobj in fdict.items():
        out[fitem] = fobj.enlarge(**bbox)

    return out


def commonBbox(fobjs):
    bbox = fobjs[0].bbox
    for fobj in fobjs[1:]:
        bbox["ymin"] = min(bbox["ymin"], fobj.bbox["ymin"])
        bbox["ymax"] = max(bbox["ymax"], fobj.bbox["ymax"])
        bbox["xmin"] = min(bbox["xmin"], fobj.bbox["xmin"])
        bbox["xmax"] = max(bbox["xmax"], fobj.bbox["xmax"])
    return bbox


def gaugeGrid(grid_template, gauge):
    out = ga.full_like(grid_template, grid_template.fill_value)
    idx = out.indexOf(gauge.y, gauge.x)
    out.data[idx] = gauge.id
    out.mask[idx] = False
    return out


def sameExtend(fobjs):
    bbox = commonBbox(fobjs)
    for fobj in fobjs:
        if fobj.bbox != bbox:
            return False
    return True


def writeReport(bpath, mask, scaling_factor):
    size = (np.sum(~mask.mask) * np.prod(np.abs(mask.cellsize))) * scaling_factor**2
    with open(os.path.join(bpath, "report.out"), "w") as f:
        f.write("calculated_catchment_size: {:}\n".format(size))


def maskData(data, mask):
    if all(x == y for x, y in zip(mask.cellsize, data.cellsize)):
        return data.setMask(mask.mask)

    enlarged_mask = mask.enlarge(**data.bbox).astype(float)
    rescaled_mask = ga.rescale(
        enlarged_mask, abs(data.cellsize[0] / mask.cellsize[0]), func="average"
    ).copy()
    return data.setMask(rescaled_mask.mask)


def main(config, gauges):

    for gauge in gauges:
        logging.info("processing gauge: %s", gauge.id)

        filedict = {}

        if not gauge.path:

            # create mask if not given
            logging.debug("reading flow accumulation")
            flowacc = ga.fromfile(config["flowacc"]).astype(np.int32)

            logging.debug("reading flow direction")
            flowdir = ga.fromfile(config["flowdir"]).astype(np.int32)

            if gauge.size:
                logging.debug("moving gauge to streamflow")
                gauge = matchFlowacc(gauge, flowacc, **config["matching"])

            if not gauge:
                warnings.warn("Failed to match the gauge to the flow accumulation grid")
                continue

            logging.debug("generating basin mask")
            mask = gaugeBasinMask(flowdir, gauge)

            # write gauge grid if desired
            if "gauge" in config:
                logging.debug("writing gauge file")
                fitem = GridFile(
                    fname=config["gauge"].get("fname", "idgauges.asc"),
                    outpath=config["gauge"].get("outpath"),
                )
                gaugefile = gaugeGrid(flowacc, gauge).shrink(**mask.bbox)
                filedict[fitem] = maskData(gaugefile, mask)

        else:
            logging.debug("reding gauge file")
            mask = gridBasinMask(gauge)

        for fdict in config["gridfiles"]:
            logging.debug("processing: %s", fdict["fname"])
            griddata = ga.fromfile(fdict["fname"]).shrink(**mask.bbox)
            filedict[GridFile(**fdict)] = maskData(griddata, mask)

        for fdict in config["ncfiles"]:
            logging.debug("processing: %s", fdict["fname"])
            fitem = NcFile(**fdict)
            ncdata = NcDimDataset(
                fitem.fname,
                "r",
                fitem.ydim,
                fitem.xdim,
                y_shift=fitem.y_shift,
                x_shift=fitem.x_shift,
            )
            filedict[NcFile(**fdict)] = maskData(ncdata.shrink(**mask.bbox), mask)

        # write mask grid if desired
        if "mask" in config:
            logging.debug("writing mask file")
            fitem = GridFile(
                fname=config["mask"].get("fname", "mask.asc"),
                outpath=config["mask"].get("outpath"),
            )
            filedict[fitem] = mask

        logging.debug("finding common extend")
        bbox = commonBbox(tuple(filedict.values()))

        logging.debug("enlarging data to common extend")
        filedict = enlargeFiles(filedict, bbox)

        if not sameExtend(tuple(filedict.values())):
            raise RuntimeError("incompatible cellsizes")

        bpath = os.path.join(config["outpath"], gauge.id)
        writeFiles(bpath, filedict)
        logging.debug("writing report")
        writeReport(bpath, mask, config["matching"]["scaling_factor"])


def initArgparser():

    parser = ArgumentParser(description="mHM basin extractor")

    parser.add_argument(
        "-n",
        "--line",
        type=int,
        help=(
            "the gauge to extract, given as its (0-based) "
            "line number in the look up table"
        ),
    )

    parser.add_argument(
        "-i",
        "--input",
        default="input.yml",
        help="the input yaml file to read (default: 'input.yml')",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        default=False,
        action="store_true",
        help="give some status output",
    )

    parser.add_argument("-c", "--cwd", default=".", help="the working directory")

    parser.add_argument("--version", action="version", version=__version__)

    return parser


if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-

import os

import numpy as np

from .netcdf4 import NcDataset


class NcDimDataset(NcDataset):
    def __init__(
        self,
        fname,
        mode="r",
        ydim="y",
        xdim="x",
        cellsize=None,
        y_shift=1,
        x_shift=0,
        *args,
        **kwargs,
    ):
        # shift_factor: the fraction of a cellsize the origin is shifted
        super(NcDimDataset, self).__init__(fname, mode, *args, **kwargs)
        self.__dict__["_y"] = ydim
        self.__dict__["_x"] = xdim
        self.__dict__["_cellsize"] = cellsize
        self.__dict__["y_shift"] = float(y_shift)
        self.__dict__["x_shift"] = float(x_shift)

    # def _delta(self):
    # ycs, xcs = np.abs(self.cellsize)
    # return ycs * self.y_shift, xcs * self.x_shift

    @property
    def bbox(self):
        x = self.variables[self._x][:]
        y = self.variables[self._y][:]
        ycs, xcs = [abs(v) for v in self.cellsize]

        return {
            "ymin": np.min(y) - (ycs * (1 - self.y_shift)),
            "ymax": np.max(y) + (ycs * self.y_shift),
            "xmin": np.min(x) - (xcs * (1 - self.x_shift)),
            "xmax": np.max(x) + (xcs * self.x_shift),
        }

    @property
    def origin(self):
        dy, dx = self.cellsize
        origin = ("l" if dy >= 0 else "u", "l" if dx >= 0 else "r")
        return "".join(origin)

    @property
    def cellsize(self):
        if self._cellsize is None:
            y = self.variables[self._y][:2]
            x = self.variables[self._x][:2]
            try:
                self.__dict__["_cellsize"] = (y[1] - y[0], x[1] - x[0])
            except IndexError:
                raise RuntimeError("Not enough coordinate values to infer cellsize")
        return self._cellsize

    @property
    def fill_values(self):
        return {k: v.fill_value for k, v in self.variables.items()}

    def shrink(self, ymin, ymax, xmin, xmax):
        def _removeCells(ymin, ymax, xmin, xmax):
            cellsize = [float(abs(v)) for v in self.cellsize]
            out = {
                "top": int(np.floor((self.bbox["ymax"] - ymax) / cellsize[0])),
                "left": int(np.floor((xmin - self.bbox["xmin"]) / cellsize[1])),
                "bottom": int(np.floor((ymin - self.bbox["ymin"]) / cellsize[0])),
                "right": int(np.floor((self.bbox["xmax"] - xmax) / cellsize[1])),
            }
            return out

        def _slices(cells):

            y = self.variables[self._y][:]
            ystart, yend = (
                (cells["bottom"], cells["top"])
                if y[0] <= y[-1]
                else (cells["top"], cells["bottom"])
            )
            yend = len(y) - yend

            x = self.variables[self._x][:]
            xstart, xend = (
                (cells["left"], cells["right"])
                if x[0] <= x[-1]
                else (cells["right"], cells["left"])
            )
            xend = len(x) - xend

            return {
                self._y: slice(ystart, yend),
                self._x: slice(xstart, xend),
            }

        cells = _removeCells(ymin, ymax, xmin, xmax)
        shrink_slices = _slices(cells)

        nc = NcDimDataset(
            None, "w", self._y, self._x, self.cellsize, self.y_shift, self.x_shift
        )
        nc.copyAttributes(self.attributes)
        nc.copyDimensions(self.dimensions, skip=(self._y, self._x))
        nc.createDimensions({k: (v.stop - v.start) for k, v in shrink_slices.items()})

        for name, var in self.variables.items():
            if any(d in var.dimensions for d in shrink_slices.keys()):
                slices = [shrink_slices.get(d, slice(None)) for d in var.dimensions]
                newvar = nc.copyVariable(var, data=False)
                newvar[:] = var[slices]
            else:
                nc.copyVariable(var, data=True)
        return nc

    def enlarge(self, ymin, ymax, xmin, xmax):
        def _padCells(ymin, ymax, xmin, xmax):
            cellsize = [float(abs(v)) for v in self.cellsize]

            out = {
                "top": int(np.ceil((ymax - self.bbox["ymax"]) / cellsize[0])),
                "left": int(np.ceil((self.bbox["xmin"] - xmin) / cellsize[1])),
                "bottom": int(np.ceil((self.bbox["ymin"] - ymin) / cellsize[0])),
                "right": int(np.ceil((xmax - self.bbox["xmax"]) / cellsize[1])),
            }
            return out

        def _slices(padding):

            y = self.variables[self._y][:]
            ystart = padding["bottom" if y[0] <= y[-1] else "top"]
            yend = ystart + len(y)

            x = self.variables[self._x][:]
            xstart = padding["left" if x[0] <= x[-1] else "right"]
            xend = xstart + len(x)

            return {self._y: slice(ystart, yend), self._x: slice(xstart, xend)}

        def _coordinates(padding):
            ycs, xcs = self.cellsize

            yvals = self.variables[self._y][:]
            xvals = self.variables[self._x][:]

            ynum = len(self.dimensions[self._y]) + padding["bottom"] + padding["top"]
            ymin = np.min(yvals) - padding["bottom"] * abs(ycs)
            ymax = np.max(yvals) + padding["bottom"] * abs(ycs)

            xnum = len(self.dimensions[self._x]) + padding["bottom"] + padding["top"]
            xmin = np.min(xvals) - padding["bottom"] * abs(xcs)
            xmax = np.max(xvals) + padding["bottom"] * abs(xcs)

            out = {
                self._y: sorted(
                    np.linspace(ymin, ymax, num=ynum, endpoint=True), reverse=ycs < 0
                ),
                self._x: sorted(
                    np.linspace(xmin, xmax, num=xnum, endpoint=True), reverse=xcs < 0
                ),
            }
            return out

        padding = _padCells(ymin, ymax, xmin, xmax)
        enlarge_slices = _slices(padding)
        coordinates = _coordinates(padding)

        nc = NcDimDataset(
            None, "w", self._y, self._x, self.cellsize, self.y_shift, self.x_shift
        )

        nc.copyAttributes(self.attributes)
        nc.copyDimensions(self.dimensions, skip=(self._y, self._x))
        nc.createDimensions({k: (len(v)) for k, v in coordinates.items()})

        for name, var in self.variables.items():
            if any(d in var.dimensions for d in enlarge_slices.keys()):
                slices = [enlarge_slices.get(d, slice(None)) for d in var.dimensions]
                newvar = nc.copyVariable(var, data=False)
                if name in (self._y, self._x):
                    newvar[:] = coordinates[name]
                else:
                    newvar[slices] = var[:]
            else:
                nc.copyVariable(var, data=True)

        return nc

    def setMask(self, mask):
        assert mask.ndim == 2, "expected a 2D mask array"

        y_idx, x_idx = np.nonzero(mask)

        mask_slices = {
            self._y: y_idx,
            self._x: x_idx,
        }

        nc = NcDimDataset(
            None, "w", self._y, self._x, self.cellsize, self.y_shift, self.x_shift
        )
        nc.copyAttributes(self.attributes)
        nc.copyDimensions(self.dimensions)

        for name, var in self.variables.items():
            newvar = nc.copyVariable(var, data=True)
            if self._y in var.dimensions and self._x in var.dimensions:
                slices = [mask_slices.get(d, slice(None)) for d in var.dimensions]
                # do it the clunky way because of a slicing bug in netCDF4
                data = newvar[:]
                data[tuple(slices)] = var.fill_value
                newvar[:] = data
        return nc

    def _header(self):
        fill_value = tuple(v for v in self.fill_values.values() if v) or (None,)
        return "\n".join(
            [
                "ncols\t{0}".format(len(self.variables[self._x])),
                "nrows\t{0}".format(len(self.variables[self._y])),
                "xllcorner\t{0}".format(self.bbox["xmin"]),
                "yllcorner\t{0}".format(self.bbox["ymin"]),
                "cellsize\t{0}".format(abs(self.cellsize[0])),
                "NODATA_value\t{0}\n".format(fill_value[0]),
            ]
        )

    def tofile(self, fname):
        with NcDataset(fname, "w") as nc:
            nc.copyDataset(self, vardata=True)
        # write a header file
        path = os.path.split(fname)[0]
        with open(os.path.join(path, "header.txt"), "w") as f:
            f.write(self._header())



# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod, abstractproperty

import numpy as np

from .utils import _broadcastTo


class _GeoBase(object):
    __metaclass__ = ABCMeta

    # @property
    # def origin(self):
    #    return "".join(
    #        ["l" if self.ycellsize > 0 else "u",
    #         "l" if self.xcellsize > 0 else "r"])

    @abstractproperty
    def coordinates(self):
        pass

    @abstractproperty
    def bbox(self):
        pass

    @abstractmethod
    def _todict(self):
        pass

    @abstractmethod
    def _replace(self):
        pass

    @abstractmethod
    def _getitem(self):
        pass

    @abstractmethod
    def toGdal(self):
        raise NotImplementedError


class _Geolocation(_GeoBase):
    def __init__(self, yvalues, xvalues, shape, origin):
        self.yvalues = yvalues
        self.xvalues = xvalues
        self.shape = shape
        self.origin = origin

    @property
    def yorigin(self):
        bbox = self.bbox
        return self.bbox["ymax" if self.origin[0] == "u" else "ymin"]

    @property
    def xorigin(self):
        return self.bbox["xmax" if self.origin[1] == "r" else "xmin"]

    @property
    def coordinates(self):
        return self.yvalues, self.xvalues

    @property
    def bbox(self):

        ymin, ymax = self.yvalues.min(), self.yvalues.max()
        xmin, xmax = self.xvalues.min(), self.xvalues.max()

        ydiff = np.abs(np.diff(self.yvalues, axis=-2))
        xdiff = np.abs(np.diff(self.xvalues, axis=-1))
        if self.origin[0] == "u":
            ymin -= ydiff[-1].max()
        else:
            ymax += ydiff[0].max()

        if self.origin[1] == "l":
            xmax += xdiff[:, -1].max()
        else:
            xmin -= xdiff[:, 0].max()

        return {"ymin": ymin, "ymax": ymax, "xmin": xmin, "xmax": xmax}

    def _replace(self, yvalues=None, xvalues=None, origin=None, shape=None):
        return _Geolocation(
            yvalues=self.yvalues if yvalues is None else yvalues,
            xvalues=self.xvalues if xvalues is None else xvalues,
            shape=self.shape if shape is None else shape,
            origin=self.origin if origin is None else origin,
        )

    def _todict(self):
        return {"yvalues": self.yvalues, "xvalues": self.xvalues}

    def _getitem(self, slc):

        yvalues = np.array(
            _broadcastTo(self.yvalues, self.shape, (-2, -1))[slc], copy=False, ndmin=2
        )

        xvalues = np.array(
            _broadcastTo(self.xvalues, self.shape, (-2, -1))[slc], copy=False, ndmin=2
        )

        if yvalues.ndim > 2:
            yvalues = yvalues[..., 0, :, :]

        if xvalues.ndim > 2:
            xvalues = xvalues[..., 0, :, :]

        return self._replace(yvalues=yvalues, xvalues=xvalues)

    def toGdal(self):
        return {
            "X_BAND": self.shape[0],  # need to be filled
            "Y_BAND": self.shape[0] + 1,  # need to be filled
            "PIXEL_OFFSET": 0,
            "LINE_OFFSET": 0,
            "PIXEL_STEP": 1,
            "LINE_STEP": 1,
        }


class _Geotrans(_GeoBase):
    def __init__(
        self, yorigin, xorigin, ycellsize, xcellsize, yparam, xparam, origin, shape
    ):
        self.yorigin = yorigin
        self.xorigin = xorigin
        self.ycellsize = ycellsize
        self.xcellsize = xcellsize
        self.yparam = yparam
        self.xparam = xparam
        self.origin = origin
        self.shape = shape
        self._yvalues = None
        self._xvalues = None

    @property
    def cellsize(self):
        return (self.ycellsize, self.xcellsize)

    @property
    def nrows(self):
        try:
            return self.shape[-2]
        except IndexError:
            return 1

    @property
    def ncols(self):
        try:
            return self.shape[-1]
        except IndexError:
            return 1

    @property
    def bbox(self):

        corners = np.array(self.getCorners())
        ymin, xmin = np.min(corners, axis=0)
        ymax, xmax = np.max(corners, axis=0)

        return {"ymin": ymin, "ymax": ymax, "xmin": xmin, "xmax": xmax}

    def _calcCoordinate(self, row, col):

        yval = self.yorigin + col * self.yparam + row * self.ycellsize
        xval = self.xorigin + col * self.xcellsize + row * self.xparam
        return yval, xval

    def toGdal(self):
        out = (
            self.xorigin,
            self.xcellsize,
            self.xparam,
            self.yorigin,
            self.yparam,
            self.ycellsize,
        )
        return out

    @property
    def coordinates(self):
        if self._yvalues is None or self._xvalues is None:
            xdata, ydata = np.meshgrid(
                np.arange(self.ncols, dtype=float), np.arange(self.nrows, dtype=float)
            )
            self._yvalues, self._xvalues = self._calcCoordinate(ydata, xdata)
        return self._yvalues, self._xvalues

    @property
    def yvalues(self):
        return self.coordinates[0]

    @property
    def xvalues(self):
        return self.coordinates[1]

    def getCorners(self):
        corners = [(0, 0), (self.nrows, 0), (0, self.ncols), (self.nrows, self.ncols)]
        return [self._calcCoordinate(*idx) for idx in corners]

    def getCorner(self, corner=None):
        if not corner:
            corner = self.origin

        bbox = self.bbox
        return (
            bbox["ymax"] if corner[0] == "u" else bbox["ymin"],
            bbox["xmax"] if corner[1] == "r" else bbox["xmin"],
        )

    def _replace(
        self,
        yorigin=None,
        xorigin=None,
        ycellsize=None,
        xcellsize=None,
        yparam=None,
        xparam=None,
        origin=None,
        shape=None,
    ):

        return _Geotrans(
            yorigin=self.yorigin if yorigin is None else yorigin,
            xorigin=self.xorigin if xorigin is None else xorigin,
            ycellsize=self.ycellsize if ycellsize is None else ycellsize,
            xcellsize=self.xcellsize if xcellsize is None else xcellsize,
            yparam=self.yparam if yparam is None else yparam,
            xparam=self.xparam if xparam is None else xparam,
            origin=self.origin if origin is None else origin,
            shape=self.shape if shape is None else shape,
        )

    def _todict(self):
        return {
            "yorigin": self.yorigin,
            "xorigin": self.xorigin,
            "ycellsize": self.ycellsize,
            "xcellsize": self.xcellsize,
            "yparam": self.yparam,
            "xparam": self.xparam,
            "origin": self.origin,
        }

    def _getitem(self, slc):

        yvalues = np.array(
            _broadcastTo(self.yvalues, self.shape, (-2, -1))[slc], copy=False, ndmin=2
        )

        xvalues = np.array(
            _broadcastTo(self.xvalues, self.shape, (-2, -1))[slc], copy=False, ndmin=2
        )

        nrows, ncols = yvalues.shape[-2:]
        ycellsize = np.diff(yvalues, axis=-2).mean() if nrows > 1 else self.ycellsize
        xcellsize = np.diff(xvalues, axis=-1).mean() if ncols > 1 else self.xcellsize

        out = self._replace(
            yorigin=yvalues.max(),
            xorigin=xvalues.min(),
            ycellsize=ycellsize,
            xcellsize=xcellsize,
            shape=yvalues.shape,
        )
        return out



# -*- coding: utf-8 -*-

from math import ceil, floor

import numpy as np


class SpatialMixin(object):
    def trim(self):
        """
        Arguments
        ---------
        None

        Returns
        -------
        GeoArray

        Purpose
        -------
        Removes rows and columns from the margins of the
        grid if they contain only fill values.
        """

        try:
            y_idx, x_idx = np.where(self.data != self.fill_value)[-2:]
            return self.removeCells(
                top=min(y_idx),
                bottom=self.nrows - max(y_idx) - 1,
                left=min(x_idx),
                right=self.ncols - max(x_idx) - 1,
            )
        except ValueError:
            return self

    def removeCells(self, top=0, left=0, bottom=0, right=0):
        """
        Arguments
        ---------
        top, left, bottom, right : int

        Returns
        -------
        GeoArray

        Purpose
        -------
        Remove the number of given cells from the respective
        margin of the grid.
        """

        top = int(max(top, 0))
        left = int(max(left, 0))
        bottom = self.nrows - int(max(bottom, 0))
        right = self.ncols - int(max(right, 0))

        return self[..., top:bottom, left:right]

    def shrink(self, ymin=None, ymax=None, xmin=None, xmax=None):
        """
        Arguments
        ---------
        ymin, ymax, xmin, xmax : scalar

        Returns
        -------
        GeoArray

        Purpose
        -------
        Shrinks the grid in a way that the given bbox is still
        within the grid domain.

        BUG:
        ------------
        For bbox with both negative and postive values
        """
        bbox = {
            "ymin": ymin if ymin is not None else self.bbox["ymin"],
            "ymax": ymax if ymax is not None else self.bbox["ymax"],
            "xmin": xmin if xmin is not None else self.bbox["xmin"],
            "xmax": xmax if xmax is not None else self.bbox["xmax"],
        }

        cellsize = [float(abs(cs)) for cs in self.cellsize]
        top = floor((self.bbox["ymax"] - bbox["ymax"]) / cellsize[0])
        left = floor((bbox["xmin"] - self.bbox["xmin"]) / cellsize[1])
        bottom = floor((bbox["ymin"] - self.bbox["ymin"]) / cellsize[0])
        right = floor((self.bbox["xmax"] - bbox["xmax"]) / cellsize[1])

        return self.removeCells(
            max(top, 0), max(left, 0), max(bottom, 0), max(right, 0)
        )

    def addCells(self, top=0, left=0, bottom=0, right=0):
        """
        Arguments
        ---------
        top, left, bottom, right : int

        Returns
        -------
        GeoArray

        Purpose
        -------
        Add the number of given cells to the respective margin of the grid.
        """

        from .core import GeoArray

        top = int(max(top, 0))
        left = int(max(left, 0))
        bottom = int(max(bottom, 0))
        right = int(max(right, 0))

        if self.origin[0] == "l":
            top, bottom = bottom, top
        if self.origin[1] == "r":
            left, right = right, left

        shape = list(self.shape)
        shape[-2:] = self.nrows + top + bottom, self.ncols + left + right

        try:
            data = np.full(shape, self.fill_value, self.dtype)
        except TypeError:
            # fill_value is set to none
            raise AttributeError(
                "Valid fill_value needed, actual value is {:}".format(self.fill_value)
            )

        yorigin = self.yorigin + top * self.ycellsize * -1
        xorigin = self.xorigin + left * self.xcellsize * -1

        out = GeoArray(
            **self._getArgs(
                data=data,
                geotrans=self.geotrans._replace(
                    yorigin=yorigin, xorigin=xorigin, shape=shape
                ),
                mode="r",
                fobj=None,
            )
        )

        # the Ellipsis ensures that the function works
        # for arrays with more than two dimensions
        out[..., top : top + self.nrows, left : left + self.ncols] = self
        return out

    def enlarge(self, ymin=None, ymax=None, xmin=None, xmax=None):
        """
        Arguments
        ---------
        ymin, ymax, xmin, xmax : scalar

        Returns
        -------
        None

        Purpose
        -------
        Enlarge the grid in a way that the given coordinates will
        be part of the grid domain. Added rows/cols are filled with
        the grid's fill value.
        """

        bbox = {
            "ymin": ymin if ymin is not None else self.bbox["ymin"],
            "ymax": ymax if ymax is not None else self.bbox["ymax"],
            "xmin": xmin if xmin is not None else self.bbox["xmin"],
            "xmax": xmax if xmax is not None else self.bbox["xmax"],
        }

        cellsize = [float(abs(cs)) for cs in self.cellsize]

        top = ceil((bbox["ymax"] - self.bbox["ymax"]) / cellsize[0])
        left = ceil((self.bbox["xmin"] - bbox["xmin"]) / cellsize[1])
        bottom = ceil((self.bbox["ymin"] - bbox["ymin"]) / cellsize[0])
        right = ceil((bbox["xmax"] - self.bbox["xmax"]) / cellsize[1])

        return self.addCells(max(top, 0), max(left, 0), max(bottom, 0), max(right, 0))

    def coordinatesOf(self, y_idx, x_idx):
        """
        Arguments
        ---------
        y_idx, x_idx :  int

        Returns
        -------
        (scalar, scalar)

        Purpose
        -------
        Return the coordinates of the grid cell definied by the given
        row and column index values. The cell corner to which the returned
        values belong is definied
        by the attribute origin:
            "ll": lower-left corner
            "lr": lower-right corner
            "ul": upper-left corner
            "ur": upper-right corner
        """

        if (y_idx < 0 or x_idx < 0) or (y_idx >= self.nrows or x_idx >= self.ncols):
            raise ValueError("Index out of bounds !")

        yorigin, xorigin = self.getCorner("ul")
        return (
            yorigin - y_idx * abs(self.cellsize[0]),
            xorigin + x_idx * abs(self.cellsize[1]),
        )

    def indexOf(self, ycoor, xcoor):
        """
        Arguments
        ---------
        ycoor, xcoor : scalar

        Returns
        -------
        (int, int)

        Purpose
        -------
        Find the grid cell into which the given coordinates
        fall and return its row/column index values.
        """

        yorigin, xorigin = self.getCorner("ul")

        yidx = int(floor((yorigin - ycoor) / float(abs(self.ycellsize))))
        xidx = int(floor((xcoor - xorigin) / float(abs(self.xcellsize))))

        if yidx < 0 or yidx >= self.nrows or xidx < 0 or xidx >= self.ncols:
            raise ValueError("Given Coordinates not within the grid domain!")

        return yidx, xidx

    # def snap(self,target):
    #     """
    #     Arguments
    #     ---------
    #     target : GeoArray

    #     Returns
    #     -------
    #     None

    #     Purpose
    #     -------
    #     Shift the grid origin that it matches the nearest cell origin in target.

    #     Restrictions
    #     ------------
    #     The shift will only alter the grid coordinates. No changes to the
    #     data will be done. In case of large shifts the physical integrety
    #     of the data might be disturbed!

    #     diff = np.array(self.getCorner()) - np.array(target.getCorner(self.origin))
    #     dy, dx = abs(diff)%target.cellsize * np.sign(diff)

    #     if abs(dy) > self.cellsize[0]/2.:
    #         dy += self.cellsize[0]

    #     if abs(dx) > self.cellsize[1]/2.:
    #         dx += self.cellsize[1]

    #     self.xorigin -= dx
    #     self.yorigin -= dy



#cython: language_level=3
# distutils: language = c++
# -*- coding: utf-8 -*-

cimport cython

import numpy as np

cimport numpy as np
from libcpp.pair cimport pair
from libcpp.stack cimport stack
from libcpp.vector cimport vector

ctypedef pair[long, long] index

cdef vector[int] Y_OFFSET = [-1, -1, -1,  0,  0,   1,  1,  1]
cdef vector[int] X_OFFSET = [-1,  0,  1, -1,  1,  -1,  0,  1]
cdef vector[int] UPSTREAM_FDIRS = [ 2,  4,  8,  1, 16, 128, 64, 32]


cpdef char[:,:] extract(int[:,:] fdir, long gauge_y, long gauge_x):
    cdef char[:,:] mask = np.zeros_like(fdir, dtype=np.int8)
    cdef stack[index] stck
    cdef index idx

    stck.push(index(gauge_y, gauge_x))
    while (not stck.empty()):
        idx = stck.top()
        stck.pop()
        mask[idx.first, idx.second] = 1
        upstreamCells(fdir, stck, idx)

    return mask

@cython.boundscheck(False)
cdef void upstreamCells(int[:, :]& fdir, stack[index]& stack, index& idx):
    cdef int nrows = fdir.shape[0]
    cdef int ncols = fdir.shape[1]
    cdef long ynn, xnn
    cdef int i
    for i in range(8):
        ynn = idx.first + Y_OFFSET[i]
        xnn = idx.second + X_OFFSET[i]
        if (ynn >= 0) and (ynn < nrows) and (xnn >= 0) and (xnn < ncols):
            if fdir[ynn, xnn] == UPSTREAM_FDIRS[i]:
                stack.push(index(ynn, xnn))






# -*- coding: utf-8 -*-

from .gdalfuncs import project, resample, rescale
from .gdalio import _DRIVER_DICT  # fromfile,
from .wrapper import (
    array,
    empty,
    fromdataset,
    fromfile,
    full,
    full_like,
    ones,
    ones_like,
    zeros,
    zeros_like,
)



# -*- coding: utf-8 -*-

import csv

import numpy as np


class Gauge(object):
    def __init__(
        self, id, y=None, x=None, size=None, path=None, varname=None, lat_fix=False
    ):
        self.id = id
        self.y = y
        self.x = x
        self.size = float(size) / np.cos(np.deg2rad(float(y))) if lat_fix else size
        self.path = path
        self.varname = varname

    def __str__(self):
        return str(
            {
                "id": self.id,
                "y": self.y,
                "x": self.x,
                "size": self.size,
                "path": self.path,
                "varname": self.varname,
            }
        )


def readGauges(fname, lat_fix=False):
    keysets = {
        tuple(sorted(["id", "size", "y", "x"])),
        tuple(sorted(["id", "path", "varname"])),
    }
    with open(fname) as f:
        reader = csv.DictReader(f, delimiter=";")
        out = []
        for gauge in reader:
            if tuple(sorted(gauge.keys())) not in keysets:
                raise RuntimeError(
                    "Header in gauges look up table must be in {:}".format(
                        list(keysets)
                    )
                )
            out.append(Gauge(lat_fix=lat_fix, **gauge))
    return out


def matchFlowacc(gauge, facc, max_distance, max_error, scaling_factor=1):
    """
    Input:
        GeoGrid or sibling, Numeric, Numeric
    Purpose:
        - Moves the Instance onto the nearest cell in the input
          flowaccumulation map (facc) based on the catchment_area.
        - The search will be done within the rectangele
          coordinate - max_distance to coordinate + max_distance
        - The method minimizes the error between the attribute
          catchment_area and the number of cells acuumulated in the
          grid. If the resulting difference is larger than
          max_error False will be returned.
        - The argument scaling factor allows to handle different
          horizontal resolutions of the information given. A factor
          to convert from map units to km^2 of the catchment area
    """
    # print "gauge before:", gauge.y, gauge.x
    y = float(gauge.y)
    x = float(gauge.x)
    size = float(gauge.size)
    bbox = {
        "ymin": y - max_distance,
        "ymax": y + max_distance,
        "xmin": x - max_distance,
        "xmax": x + max_distance,
    }
    grid = facc.shrink(**bbox).astype("float32")
    grid *= abs(
        (grid.cellsize[0] * scaling_factor) * (grid.cellsize[1] * scaling_factor)
    )

    gauge_y_idx, gauge_x_idx = grid.indexOf(y, x)

    # find all possible river cells in the nearest possible distance
    error = 0
    river_cells_y = []
    while len(river_cells_y) == 0 and error < max_error:
        river_cells_y, river_cells_x = np.where(
            (grid.data >= size * (1 - error)) & (grid.data <= size * (1 + error))
        )
        error += 0.01

    if error <= max_error:
        # the closest river cell
        nn = (
            (river_cells_y - gauge_y_idx) ** 2 + (river_cells_x - gauge_x_idx) ** 2
        ).argmin()

        # the cell cordinates
        y, x = grid.coordinatesOf(river_cells_y[nn], river_cells_x[nn])
        return Gauge(id=gauge.id, y=y, x=x, size=size)



# -*- coding: utf-8 -*-

import os
import warnings

import numpy as np

try:
    from osgeo import gdal, osr
except ImportError:
    import gdal, osr

from .gdalspatial import _Projection
from .geotrans import _Geotrans

gdal.UseExceptions()
gdal.PushErrorHandler("CPLQuietErrorHandler")

# should be extended, for available options see:
# http://www.gdal.org/formats_list.html
_DRIVER_DICT = {
    ".tif": "GTiff",
    ".asc": "AAIGrid",
    ".img": "HFA",
    ".sdat": "SAGA",
    ".png": "PNG",  # not working properly
}

# type mapping:
#     - there is no boolean data type in GDAL
_TYPEMAP = {
    "uint8": 1,
    "int8": 1,
    "uint16": 2,
    "int16": 3,
    "uint32": 4,
    "int32": 5,
    # "int64"      : 5, # there is no int64 data type in GDAL, map to int32 and issue a warning
    "float32": 6,
    "float64": 7,
    "complex64": 10,
    "complex128": 11,
    1: "int8",
    2: "uint16",
    3: "int16",
    4: "uint32",
    5: "int32",
    6: "float32",
    7: "float64",
    10: "complex64",
    11: "complex128",
}

_COLOR_DICT = {
    1: "L",
    2: "P",
    3: "R",
    4: "G",
    5: "B",
    6: "A",
    7: "H",
    8: "S",
    9: "V",
    10: "C",
    11: "M",
    12: "Y",
    13: "K",
    14: "Y",
    15: "Cb",
    16: "Cr",
}

_COLOR_MODE_LIST = ("L", "P", "RGB", "RGBA", "CMYK", "HSV", "YCbCr")

_FILE_MODE_DICT = {"r": gdal.GA_ReadOnly, "v": gdal.GA_ReadOnly, "a": gdal.GA_Update}


def _fromFile(fname, mode="r"):
    """
    Parameters
    ----------
    fname : str  # file name

    Returns
    -------
    GeoArray

    Purpose
    -------
    Create GeoArray from file

    """

    if mode not in _FILE_MODE_DICT:
        raise TypeError(
            "Supported file modes are: {:}".format(", ".join(_FILE_MODE_DICT.keys()))
        )

    fobj = gdal.OpenShared(fname, _FILE_MODE_DICT[mode])
    if fobj:
        return _fromDataset(fobj, mode)
    raise IOError("Could not open file: {:}".format(fname))


def _getColorMode(fobj):
    tmp = []
    for i in range(fobj.RasterCount):
        color = fobj.GetRasterBand(i + 1).GetColorInterpretation()
        tmp.append(_COLOR_DICT.get(color, "L"))
    return "".join(sorted(set(tmp), key=tmp.index))


def _fromDataset(fobj, mode="r"):

    from .core import GeoArray

    def _parseGeotrans(geotrans):
        return {
            "yorigin": geotrans[3],
            "xorigin": geotrans[0],
            "ycellsize": geotrans[5],
            "xcellsize": geotrans[1],
            "origin": "ul",  # is that always true?
            "yparam": geotrans[4],
            "xparam": geotrans[2],
        }

    fill_values = tuple(
        fobj.GetRasterBand(i + 1).GetNoDataValue() for i in range(fobj.RasterCount)
    )

    if len(set(fill_values)) > 1:
        warnings.warn(
            "More then on fill value found. Only {:} will be used".format(
                fill_values[0]
            ),
            RuntimeWarning,
        )

    data = fobj.GetVirtualMemArray() if mode == "v" else fobj.ReadAsArray()
    # NOTE: not to robust...
    geotrans = _Geotrans(shape=data.shape, **_parseGeotrans(fobj.GetGeoTransform()))

    return GeoArray(
        data=data,
        fill_value=fill_values[0],
        proj=_Projection(fobj.GetProjection()),
        mode=mode,
        color_mode=_getColorMode(fobj),
        fobj=fobj,
        geotrans=geotrans,
    )


def _getDataset(grid, mem=False):

    # Returns an gdal memory dataset created from the given grid

    if grid._fobj and not mem:
        return grid._fobj

    driver = gdal.GetDriverByName("MEM")

    try:
        out = driver.Create(
            "", grid.ncols, grid.nrows, grid.nbands, _TYPEMAP[str(grid.dtype)]
        )
    except KeyError:
        raise RuntimeError("Datatype {:} not supported by GDAL".format(grid.dtype))

    if isinstance(grid.geotrans, _Geotrans):
        out.SetGeoTransform(grid.toGdal())

    if grid.proj:
        out.SetProjection(grid.proj.toWkt())

    for n in range(grid.nbands):
        band = out.GetRasterBand(n + 1)
        if grid.fill_value is not None:
            band.SetNoDataValue(float(grid.fill_value))
        data = grid[n] if grid.ndim > 2 else grid
        band.WriteArray(data)

    return out


def _toFile(geoarray, fname):
    """
    Arguments
    ---------
    fname : str  # file name

    Returns
    -------
    None

    Purpose
    -------
    Write GeoArray to file. The output dataset type is derived from
    the file name extension. See _DRIVER_DICT for implemented formats.
    """

    def _fnameExtension(fname):
        return os.path.splitext(fname)[-1].lower()

    def _getDriver(fext):
        """
        Guess driver from file name extension
        """
        if fext in _DRIVER_DICT:
            driver = gdal.GetDriverByName(_DRIVER_DICT[fext])
            metadata = driver.GetMetadata_Dict()
            if "YES" == metadata.get("DCAP_CREATE", metadata.get("DCAP_CREATECOPY")):
                return driver
            raise IOError("Datatype cannot be written")
        raise IOError("No driver found for filename extension '{:}'".format(fext))

    def _getDatatype(driver):
        tnames = tuple(driver.GetMetadata_Dict()["DMD_CREATIONDATATYPES"].split(" "))
        types = tuple(gdal.GetDataTypeByName(t) for t in tnames)
        tdict = tuple((gdal.GetDataTypeSize(t), t) for t in types)
        otype = max(tdict, key=lambda x: x[0])[-1]
        return np.dtype(_TYPEMAP[otype])

    dataset = _getDataset(geoarray)
    driver = _getDriver(_fnameExtension(fname))
    driver.CreateCopy(fname, dataset, 0)


def _writeData(grid):

    fobj = grid.fobj
    data = grid.data
    if data.ndim == 2:
        data = data[None, ...]

    for n in range(fobj.RasterCount):
        fobj.GetRasterBand(n + 1).WriteArray(data[n])



# -*- coding: utf-8 -*-


import numpy as np

try:
    basestring
except NameError:
    basestring = str


def _broadcastTo(array, shape, dims):
    """
    array, shape: see numpy.broadcast_to
    dims: tuple, the dimensions the array dimensions
          should end up in the output array
    """
    assert len(array.shape) == len(dims)
    assert len(set(dims)) == len(dims)  # no duplicates
    # handle negative indices
    dims = [d if d >= 0 else d + len(shape) for d in dims]
    # bring array to the desired dimensionality
    slc = [slice(None, None, None) if i in dims else None for i in range(len(shape))]
    return np.broadcast_to(array[tuple(slc)], shape)


def _broadcastedMeshgrid(*arrays):
    def _toNd(array, n, pos=-1):
        """
        expand given 1D array to n dimensions,
        the dimensions > 0 can be given by pos
        """
        assert array.ndim == 1, "arrays should be 1D"
        shape = np.ones(n, dtype=int)
        shape[pos] = len(array)
        return arr.reshape(shape)

    shape = tuple(len(arr) for arr in arrays)

    out = []
    for i, arr in enumerate(arrays):
        tmp = np.broadcast_to(_toNd(arr, len(shape), pos=i), shape)
        # there should be a solution without transposing...
        out.append(tmp.T)
    return out


def _tupelize(arg):
    out = arg
    if isinstance(out, basestring):
        out = (out,)
    if not isinstance(out, tuple):
        try:
            out = tuple(out)
        except TypeError:
            out = (out,)
    return out



# -*- coding: utf-8 -*-
"""Basin Extractor."""
import os
import site
import sys

import numpy as np
from Cython.Build import cythonize
from setuptools import Extension, setup

# pip issue: https://github.com/pypa/pip/issues/7953
site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

# cython extensions ###########################################################

CY_MODULES = []
CY_MODULES.append(
    Extension(
        "basinex.extractor",
        [os.path.join("src", "basinex", "extractor.pyx")],
        language="c++",
        include_dirs=[np.get_include()],
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    )
)
EXT_MODULES = cythonize(CY_MODULES)  # annotate=True

# embed signatures for sphinx
for ext_m in EXT_MODULES:
    ext_m.cython_directives = {"embedsignature": True}

# setup #######################################################################

setup(ext_modules=EXT_MODULES, include_dirs=[np.get_include()])



# -*- coding: utf-8 -*-
# from collections import namedtuple


class NcFile(object):
    def __init__(self, fname, ydim, xdim, outpath=None, y_shift=1, x_shift=0):
        self.fname = fname
        self.ydim = ydim
        self.xdim = xdim
        self.outpath = outpath
        self.y_shift = y_shift
        self.x_shift = x_shift


class GridFile(object):
    def __init__(self, fname, outpath=None):
        self.fname = fname
        self.outpath = outpath
